"""Custom exceptions for phrases module."""


class PhrasesError(Exception):
    """Custom Error for Phrase Trees."""

    def __init__(self, value: str, message: str) -> None:
        """Construct exception."""
        self.value = value
        self.message = message
        super().__init__()



"""Optom Automation.

Date: 7/07/2022
"""
from collections.abc import Iterable

from rich.align import Align
from rich.console import Console, Group
from rich.panel import Panel
from rich.prompt import IntPrompt
from rich.text import Text
from rich.theme import Theme

from optom_automator.__init__ import __app__, __author__, __version__
from optom_automator.phrases import Phrases
from optom_automator.reader import Reader

reader = Reader()
phrase = Phrases()

custom_theme = Theme(
    {
        "menu_title": "bold cyan",
        "menu_item": "bold magenta",
        "dim": "dim",
        "number": "green",
    }
)
console = Console(theme=custom_theme)


def _detail_element(text: str) -> Text:
    t = Text(text, style="dim")
    t.highlight_regex(r":\s\w+", style="number")
    return t


def _menu_element(text: str) -> Text:
    t = Text(text, style="normal")
    t.highlight_regex(r"\d+\.", style="menu_item")
    t.highlight_regex(r"'\w+'", style="number")
    return t


def _voice_details() -> list[Align]:
    voice_info = reader.voices[reader.voice]
    elements = [
        Align.left(_detail_element(f"\nVoice: {voice_info.name}")),
        Align.left(_detail_element(f"Rate: {reader.rate} words/min")),
    ]
    return elements


def generate_menu(
    choices: Iterable[str | None],
    menu_query: str = "Please choose an option:",
    default_choice: int = 0,
    padding: int = 4,
) -> int:
    """Generate the menu. And also ask for user input.

    Args:
        choices (Iterable[str | None]): Options in the menu.
        menu_query (str): What the question test will be displayed.
        default_choice (int): The default choice as an index.
        padding (int): The padding for the display.

    Returns:
        int: The choice by the user.
    """
    formatted = [
        Align.left(_menu_element(f"{i+1}. {choice}"))
        for i, choice in enumerate(choices)
    ]
    menu_formatted = Align.center(menu_query + "\n", style="menu_title")
    groups = [menu_formatted] + formatted + _voice_details()
    panel = Panel.fit(
        Align.left(Group(*groups)),
        title=f"{__app__} v{__version__}",
        subtitle=f"By {__author__}",
        padding=padding,
    )
    console.clear()
    console.print(panel)
    idx = [str(i + 1) for i, _ in enumerate(choices)]
    choice = IntPrompt.ask(
        choices=idx,
        default=idx[default_choice],
    )
    return int(choice)


def main() -> None:
    """Run main."""
    with console.screen(hide_cursor=False):
        while True:
            with console.status("Reading...", spinner="aesthetic"):
                reader.phrase = phrase.read_phrase()
                reader.read()
            choices = [
                f"Repeat - '{phrase.current}'",
                f"Next - '{phrase.peek}'",
                "Repeat (slower)",
                "Repeat (faster)",
                "Previous",
                "Exit",
            ]
            choice = generate_menu(choices=choices)
            match choice:
                case 1:  # Repeat
                    pass
                case 2:  # Next
                    phrase.next()
                case 3:  # Repeat slower
                    reader.rate -= 50
                case 4:  # Repeat faster
                    reader.rate += 50
                case 5:  # Previous
                    phrase.previous()
                case 6:  # Exit
                    break


if __name__ == "__main__":
    main()



"""Custom logging."""

import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
LEVEL = logging.INFO

logging.basicConfig(
    level=LEVEL,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler()],
)

log = logging.getLogger("rich")



"""The Phrase Tree."""

from .types import StageType

PHRASE_TREE: dict[str, StageType] = {
    "GREETING": ("random", ("Kia ora", "Hello", "Welcome")),
    "HISTORY": (
        "random",
        (
            "How can I help you out today with your eyesight?",
            "So, how is everything getting along with your vision?",
        ),
    ),
    "RIGHT_VA": (
        "random",
        (
            "Let's take a look at how well you can see. Starting with your right eye, what is the smallest line that you can read?",
        ),
    ),
    "LEFT_VA": (
        "random",
        ("Now with your left eye, what is the smallest line that you can read?",),
    ),
    "RIGHT_REDGREEN": (
        "random",
        (
            "Is it clearer on the red or green side?",
            "Are the symbols clearer on the red or green square?",
        ),
    ),
    "RIGHT_CYLINDER": (
        "sequence",
        (
            "Is it clearer with view 1 or view 2 or about the same?",
            "Is it clearer with lens 3 or lens 4 or about the same?",
            "Is it clearer with lens 5 or lens 6 or about the same?",
            "Is it clearer with lens 7 or lens 8 or about the same?",
            "Is it clearer with lens 9 or lens 10 or about the same?",
        ),
    ),
    "RIGHT_SPHERE": (
        "random",
        ("Is it clearer of worse with the next lens or about the same?",),
    ),
    "LEFT_REDGREEN": (
        "random",
        (
            "Let's do your left eye, is it clearer on the red or green side?",
            "Are the symbols clearer on the red or green square?",
        ),
    ),
    "LEFT_CYLINDER": (
        "sequence",
        (
            "Is it clearer with view 1 or view 2 or about the same?",
            "Is it clearer with lens 3 or lens 4 or about the same?",
            "Is it clearer with lens 5 or lens 6 or about the same?",
            "Is it clearer with lens 7 or lens 8 or about the same?",
            "Is it clearer with lens 9 or lens 10 or about the same?",
        ),
    ),
    "LEFT_SPHERE": (
        "random",
        ("Is it clearer of worse with the next lens or about the same?",),
    ),
    "END_EXAM": (
        "random",
        (
            "You have done well. Thank you.",
            "This concludes the lenses part of the eye exam.",
        ),
    ),
}



"""Exception for Reader module."""


class ReaderError(Exception):
    """Exception for invalid ReadInstruction."""

    def __init__(self, message):
        """Construct class."""
        self.message = message
        super().__init__()



"""Contains the instruction class."""

import pyttsx3

from .exceptions import ReaderError


class Reader:
    """Text to speech class wrapper.

    Attributes:
        phrase (str): The phrase to be uttered.
        rate (int): The rate of utterance.
        volume (float): The volumn of utterance.
        voice (int): The voice of utterance.
        voices (list): The voices available to set for utterance.

    Methods:
        read(): Reads out the phrase with settings.
    """

    def __init__(
        self,
        phrase: str = "",
        rate: int = 200,
        volume: float = 1.00,
        voice: int = 0,
    ):
        """Construct the class.

        Args:
            phrase (str): Phrase to be uttered.
            rate (int): Rate defined by pyttsx3.
            volume (float): Volume defined by pyttsx3.
            voice (int): The voice defined by pyttsx3 and the OS.
        """
        self._phrase = phrase
        self._rate = rate
        self._volume = volume
        self._voice = voice

        self._engine = pyttsx3.init()

    @property
    def phrase(self) -> str:
        """Return the phrase.

        Returns:
            (str): The phrase to be uttered.

        Examples:
            >>> instruction.phrase
            "A phrase"
        """
        return self._phrase

    @phrase.setter
    def phrase(self, phrase: str) -> None:
        """Set phrase to be uttered.

        If the phrase is out of range, the property will be set to False.

        Args:
            phrase (int): Phrase index.

        Examples:
            >>> instruction.phrase = "Hello"
            >>> instruction.phrase
            "Hello"
        """
        self._phrase = phrase

    @property
    def rate(self) -> int:
        """Get the rate of utterance.

        Returns:
            int: The rate of the phase.

        Examples:
            >>> instruction.rate
            200
        """
        return self._rate

    @rate.setter
    def rate(self, rate: int) -> None:
        """Set the rate of utterance.

        Args:
            rate (int): The speed of utterance.

        Examples:
            >>> instruction.rate = 50
            >>> instruction.rate
            50
        """
        self._rate = rate

    @property
    def volume(self) -> float:
        """Return the volume setting.

        Returns:
            float: The current volume setting.
        """
        return self._volume

    @volume.setter
    def volume(self, volume: float) -> None:
        """Set the volume.

        Args:
            volume (float): The volume setting. 1.0 being 100% volume.
        """
        self._volume = volume

    @property
    def voices(self) -> list:
        """Return a list of voices depending on the OS.

        Returns:
            list: Voices available.
        """
        return self._engine.getProperty("voices")

    @property
    def voice(self) -> int:
        """Return current voice by index.

        Returns:
            int: The voice by index.
        """
        return self._voice

    @voice.setter
    def voice(self, voice: int) -> None:
        """Set the voice by index.

        If out of index, the voice will be set to the last indexed voice.

        Args:
            voice (int): [TODO:description]
        """
        voice = abs(int(voice))
        self._voice = voice
        if voice > len(self.voices) - 1:
            self._voice = len(self.voices) - 1

    def read(self):
        """Read the stored phrase.

        Raises:
            InvalidReadInstruction: This is raise if the phrase is empty or set
            to None.
        """
        if self.phrase == "" or self.phrase is None:
            raise ReaderError(message="No phrase set.")
        self._engine.say(self.phrase)
        self._engine.setProperty("rate", self.rate)
        self._engine.setProperty("volume", self.volume)
        self._engine.setProperty("voice", self.voices[self.voice].id)
        self._engine.runAndWait()
        self._engine.stop()

    def __repr__(self):
        """Representation of object."""
        cls = self.__class__.__name__
        return f"{cls}(phrase={self.phrase}, rate={self.rate}, volume={self.volume}, voice={self.voice})"

    def __str__(self):
        """Stringfy object."""
        return self._phrase



"""Phrases module."""

import random

from .exceptions import PhrasesError
from .phrase_tree import PHRASE_TREE
from .types import StageType


class Phrases:
    """Interface to allow phrase tree to be access.

    Attributes:
        current (str): The current stage.
        peek (str): Provides the next stage.

    Methods:
        read_phrase(): Reads phrase from stage.
        next(): Moves to the next stage.
    """

    def __init__(self, phrase_tree: dict[str, StageType] = PHRASE_TREE):
        """Construct Phrases object.

        Args:
            phrase_tree (dict[str, StageType]): This is the phrase tree,
            containing all dialogue.
        """
        self._current = 0
        self._phrase_tree = phrase_tree
        self._stages = list(phrase_tree.keys())
        self._sequence = 0

    @property
    def current(self) -> str:
        """Provide current stage. And not a index.

        Returns:
            str: The current stage.

        Examples:
            >>> phrase.current_phrase
            >>> "This is the current phrase"
        """
        return self._stages[self._current]

    @property
    def peek(self) -> str | None:
        """Return the next stage by peeking.

        If the current stage is the last stage. Then `None` is returned.

        Returns:
            str | None: The next phrase. Or False if there are no more phrases.

        Examples:
        >>> phrase.peek
        "The next phrase"

        # example when current phrase is last
        >>> phrase.peek
        None
        """
        next_idx = self._current + 1
        if next_idx > len(self._stages) - 1:
            return None
        return self._stages[next_idx]

    def _reset_sequence(self, default: int = 0):
        self._sequence = default

    def next(self) -> bool:
        """Move to next stage.

        Returns:
            bool: To confirm if the move took place. `True` if it moved to next
            stage and `False` if it did not (i.e. at last stage).

        Examples:
            >>> phrase.read_phrase()
            "Current phrase."
            >>> phrase.next()
            True
            >>> phrase.read_phrase()
            "Next and final pharse."
            >>> phrase.next()
            False
        """
        if self.peek is None:
            return False
        self._reset_sequence()
        self._current += 1
        return True

    def previous(self) -> bool:
        """Move to previous stage.

        Returns:
            bool: To confirm is moved to previous stage. `True` if successful,
            and `False` is not (i.e. at first stage).

        Examples:
            >>> phrase.read_phrase()
            "2nd phrase."
            >>> phrase.previous()
            True
            >>> phrase.read_phrase()
            "1st phrase."
            >>> phrase.previous()
            False
        """
        if self._current > 0:
            self._reset_sequence()
            self._current -= 1
            return True
        return False

    def _read_random(self, phrases):
        phrases = self._phrase_tree[self.current][1]
        return random.choice(phrases)

    def _read_sequence(self, phrases):
        phrase = phrases[self._sequence]
        self._sequence += 1
        if self._sequence > len(phrases) - 1:
            self._reset_sequence()
        return phrase

    def read_phrase(self) -> str:
        """Read the phrase according to the stage.

        Returns:
            str: The phrase.

        Examples:
        >>> phrase.read_phrase()
        "Reading phrase"
        """
        stage = self._phrase_tree[self.current]
        match stage[0]:
            case "random":
                phrase = self._read_random(stage[1])
            case "sequence":
                phrase = self._read_sequence(stage[1])
            case _:
                raise PhrasesError(
                    value=stage[0],
                    message="The stage types can only be set to 'random' and 'sequence'.",
                )
        return phrase

    def __len__(self) -> int:
        """Return number of phrases.

        Returns:
            int: Number of phrases.
        """
        return len(self._stages)

    def __repr__(self) -> str:
        """Representation of the object.

        Returns:
            str: The object representation.
        """
        return "phrase_object"

    def __str__(self) -> str:
        """Call string of object.

        Returns:
            str: The phrase the object is set to. As well as the index and
            total size.
        """
        return (
            f"({self._current}/{len(self._phrase_tree)}): {self._stages[self._current]}"
        )
